import numpy as np
import pennylane as qml
import torch


class QuantumFCLayer:
    """
    Quantum fully-connected layer for hybrid quantum-classical neural networks.

    This class defines a quantum layer that applies various quantum encodings
    and ansatz circuits to process input data using a quantum neural network.

    Attributes:
        input_size (int): Dimension of the input data.
        n_layers (int): Number of layers in the quantum ansatz circuit.
        encoding (str): Name of the data encoding method. 
            Supported options: 'qubit', 'amplitude', 'ZZFeatureMap', 'QAOA'.
        ansatz (int): Identifier for the ansatz quantum circuit (1-6).
        observables (str or list of str): Observables measured at the end of the circuit.
            If `None`, defaults to 'Z' * nqbits.
        backend (str): Quantum backend used for execution.
            Defaults to 'default.qubit' (Pennylane simulator).
        nqbits (int): Number of qubits required for encoding.
        dev (qml.Device): Quantum device used for simulation.
    """

    def __init__(self, input_size, n_layers=2, encoding='qubit', ansatz=1,
                 observables=None, backend="default.qubit"):
        """
        Initializes the quantum fully-connected layer.

        Args:
            input_size (int): Dimension of the input data.
            n_layers (int, optional): Number of layers in the quantum ansatz circuit.
                Defaults to 2.
            encoding (str, optional): Encoding method for quantum data. 
                Supported options: 'qubit', 'amplitude', 'ZZFeatureMap', 'QAOA'. Defaults to 'qubit'.
            ansatz (int, optional): Identifier for the quantum ansatz circuit (1-6). Defaults to 1.
            observables (str or list, optional): Observables measured at the end of the circuit. 
                If `None`, defaults to 'Z' * nqbits.
            backend (str, optional): Quantum backend for execution. 
                Defaults to 'default.qubit' (Pennylane simulator).

        Raises:
            NotImplementedError: If an invalid ansatz number is provided.
        """
        self.input_size = input_size
        self.n_layers = n_layers
        self.encoding = encoding

        # Calculate the number of qubits needed for the data encoding
        if self.encoding == 'amplitude':
            self.nqbits = int(np.ceil(np.log2(self.input_size)))
        else:
            self.nqbits = self.input_size

        ansatz_names = [
            "circuit_10", "circuit_9", "circuit_15", "circuit_14",
            "circuit_13", "circuit_6"
        ]
        if type(ansatz) != int or ansatz < 1 or ansatz > 6:
            raise NotImplementedError(
                'Choose a quantum ansatz between 1 and 6')
        self.ansatz = ansatz_names[ansatz-1]
        self.get_shapes()

        self.observables = \
            "Z"*self.nqbits if observables is None else observables
        
        if isinstance(self.observables, list): # Convert to operators
            self.observables = [qml.pauli.string_to_pauli_word(o) for o in self.observables]
        self.backend = backend
        self.dev = qml.device(self.backend, wires=self.nqbits)

    ###########################################################################
    # Quantum encodings
    ###########################################################################

    def _qubit_encoding(self, nqbits, feature_vector):
        """
        Generates the circuit that performs qubit encoding.

        Args:
            nqbits (int): Number of qubits
            feature_vector (array): input data
        """
        qml.AngleEmbedding(features=feature_vector, wires=range(nqbits), rotation='X')


    def _ZZFeatureMap_encoding(self, nqbits, feature_vector, n_layers=2):
        """
        Generates the circuit that performs ZZFeatureMap encoding.

        Args:
            nqbits (int): Number of qubits
            feature_vector (array): Input data
            n_layers (int): Number of layers
        """
        for _ in range(n_layers):
            for i in range(nqbits):
                qml.Hadamard(wires = i)
                qml.U1(2*feature_vector[i], wires =i)
            # CNOT + hadamard
            for i in range(1,nqbits):
                if i-2>=0:
                    qml.CNOT(wires=[i-2, i-1])
                qml.CNOT(wires=[i-1, i])
                qml.U1(2*(np.pi - feature_vector[i-1])*(np.pi - feature_vector[i]), wires = i)
            qml.CNOT(wires=[nqbits - 2, nqbits-1])

    def _amplitude_encoding(self, feature_vector):
        """
        Generates the circuit that performs amplitude encoding.

        Args:
            feature_vector (array): Input data
        """
        qml.AmplitudeEmbedding(features=feature_vector, wires=range(self.nqbits),normalize=True)

    def _QAOA_encoding(self, feature_vector, input_weights):
        """
        Generates the circuit that performs QAOA encoding. Notice that this encoding has trainable parameters

        Args:
            feature_vector (array): Input data
            input_weights (array): shape (L,1) for 1 qubit, (L,3) for two qubits and (L, 2*nqbits) otherwise
        """
        qml.QAOAEmbedding(features=feature_vector, weights=input_weights, wires=range(len(feature_vector)))

    ###########################################################################
    # Quantum ansatz
    ###########################################################################

    def _circuit_10(self, qubits, weights):

        """
        Generates the quantum circuit corresponding to the circuit_10 ansatz
        (option 1)

        Args:
            qbits (int): Qubits to apply this circuit
            weights (array): shape (n_layers, nqbits, 3).
        """
        # Choose the number of layers of the PQC
        qml.StronglyEntanglingLayers(weights=weights, wires=qubits)

    def _circuit_6(self, qubits, weights):
        """
        Generates the quantum circuit corresponding to the circuit_6 ansatz
        (option 6)

        Args:
            qbits (int): Qubits to apply this circuit
            weights (array): shape (nqbits, [ 1 (Rx1) + 1 (Rz1) + 1 x nqbits -1 (cRx) + 1 (Rx2) + 1 (Rz2)] x n_layers)
        """
        size = weights.shape[1]
        nqbits = len(qubits)
        n_layers = int(size/(4 + (nqbits - 1)))

        for layer in range(n_layers):
            # First Rx layer
            for i in range(nqbits):
                qml.RX(weights[i,0 + layer-1], wires = qubits[i])
            # First RZ layer
            for i in range(nqbits):
                qml.RZ(weights[i,1 + layer-1], wires = qubits[i])

            # Controlled Rx
            l_idx = 0
            for i in range(nqbits-1, -1, -1):  # i = controlled qubits
                for q in range(nqbits-1, -1, -1):
                    if i == q:
                        l_idx += 1
                        # Can't link to itself
                        continue
                    ly = l_idx//nqbits
                    lx = l_idx%nqbits
                    qml.CRX(weights[lx, 1 + ly + layer-1], wires = [qubits[i], qubits[q]])
                    l_idx += 1

            # Second Rx layer
            for i in range(nqbits):
                qml.RX(weights[i,4 + nqbits - 3 + layer-1], wires = qubits[i])
            # Second Ry layer
            for i in range(nqbits):
                qml.RZ(weights[i,4 + nqbits - 2 + layer-1], wires = qubits[i])


    def _circuit_9(self, qubits, weights):
        """
        Generates the quantum circuit corresponding to the circuit_9 ansatz
        (option 2)

        Args:
            qbits (int): Qubits to apply this circuit
            weights (array): shape (nqbits,  n_layers)
        """
        n_layers = weights.shape[1]
        nqbits = len(qubits)
        qbit_list = [(i, i-1) for i in range(nqbits-1, 0, -1)]

        for layer in range(n_layers):

            # 1. Hadamard
            for q in range(nqbits):
                qml.Hadamard(wires = qubits[q])

            # 2. Phase shifts
            for (qbit1, qbit2) in qbit_list:
                qml.CZ(wires = [qubits[qbit1], qubits[qbit2]])

            # 3. Ry gates
            for q in range(nqbits):
                qml.RX(weights[q, layer], wires = qubits[q])


    def _circuit_15(self, qubits, weights):
        """
        Generates the quantum circuit corresponding to the circuit_15 ansatz
        (option 3)

        Args:
            qbits (int): Qubits to apply this circuit
            weights (array): shape (n_layers, nqbits)
        """
        qml.BasicEntanglerLayers(weights=weights, wires=qubits)

    def _circuit_14(self, qubits, weights):
        """
        Generates the quantum circuit corresponding to the circuit_14 ansatz
        (option 4)

        Args:
            qbits (int): Qubits to apply this circuit
            weights (array): shape (nqbits ,  [1 (Ry1) + 1 (CRx1) + 1 (Ry2) + 1 (CRx2)]*n_layers)
        """
        n_layers = int(weights.shape[1]/4)
        nqbits = len(qubits)
        qbit_list = [(0, nqbits-1)] + [(i,i-1) for i in range(nqbits-1,0,-1)]
        qbit_list2 = [(nqbits-2, nqbits-1), (nqbits-1,0)] + [(i,i+1) for i in range(0, nqbits-2)]
        for layer in range(n_layers):
            # 1. Ry gates
            for i in range(nqbits):
                qml.RY(weights[i,0 + layer -1], wires = qubits[i])

            # 2. Controlled Rx
            for (qbit1, qbit2) in qbit_list:
                qml.CRX(weights[i,1 + layer -1], wires = [qubits[qbit2], qubits[qbit1]])

            # 3. Ry gates
            for i in range(nqbits):
                qml.RY(weights[i,2 + layer -1], wires = qubits[i])

            # 4. Controlled Rx
            for (qbit1, qbit2) in qbit_list2:
                qml.CRX(weights[i,1 + layer -1], wires = [qubits[qbit2], qubits[qbit1]])


    def _circuit_13(self, qubits, weights):
        """
        Generates the quantum circuit corresponding to the circuit_9 ansatz
        (option 5)

        Args:
            qbits (int): Qubits to apply this circuit
            weights (array): shape (nqbits x [1 (Ry1) + 1 (CRx1) + 1 (Ry2) + 1 (CRx2)]*n_layers)
        """
        n_layers = int(weights.shape[1]/4)
        nqbits = len(qubits)
        qbit_list = [(0, nqbits-1)] + [(i,i-1) for i in range(nqbits-1,0,-1)]
        qbit_list2 = [(nqbits-2, nqbits-1), (nqbits-1,0)] + [(i,i+1) for i in range(0, nqbits-2)]
        for layer in range(n_layers):
            # 1. Ry gates
            for i in range(nqbits):
                qml.RY(weights[i,0 + layer -1], wires = qubits[i])

            # 2. Controlled Rx
            for (qbit1, qbit2) in qbit_list:
                qml.CRZ(weights[i,1 + layer -1], wires = [qubits[qbit2], qubits[qbit1]])

            # 3. Ry gates
            for i in range(nqbits):
                qml.RY(weights[i,2 + layer -1], wires = qubits[i])

            # 4. Controlled Rx
            for (qbit1, qbit2) in qbit_list2:
                qml.CRZ(weights[i,1 + layer -1], wires = [qubits[qbit2], qubits[qbit1]])
    def apply_ansatz(self, weights_layer, qubits = None):
        """
        Applies the quantum ansatz to the quantum neural network.

        Args:
            weight_layers (array): Weights of the ansatz layer
            qubits: qubits to apply the ansatz to

        Returns:
            (QuantumNN): quantum neural network
        """
        # Ansatz
        name_to_func = {
            "circuit_10": self._circuit_10,
            "circuit_6": self._circuit_6,
            "circuit_9": self._circuit_9,
            "circuit_15": self._circuit_15,
            "circuit_14": self._circuit_14,
            "circuit_13": self._circuit_13,
        }
        if self.ansatz not in name_to_func:
            raise NotImplementedError(
                f"Quantum Circuit model '{self.ansatz}' not implemented. "
                "Implemented models: " +
                ", ".join(list(name_to_func.keys()))
            )
        if qubits is None:
            qubits = range(self.nqbits)
        # Apply anstatz
        name_to_func[self.ansatz](qubits, weights_layer)

        if self.observables!=False:
            # Return measurements
            if self.observables=='probs':  # Case 1: Measuring probabilities
                return qml.probs(wires=list(range(self.nqbits)))
            elif self.observables=='state':  # Case 2: Measuring states
                return qml.state()
            else: # Case 3: Measuring observables
                return [qml.expval(obs) for obs in self.observables]


    def qnn_layer(self,inputs, weights_layers):
        """
        Creates the quantum neural network composed of the quantum encoding with qubit, amplitude or ZZ encoding, 
        and a quantum ansatz.

        Args:
            inputs (array): Input data
            weight_layers (array): Weights of the ansatz layer

        Returns:
            (QuantumNN): quantum neural network
        """
        # Data encoding
        if self.encoding == 'angle' or self.encoding=='qubit':
            self._qubit_encoding(self.nqbits, inputs)
        elif self.encoding == 'amplitude':
            self._amplitude_encoding(inputs)
        elif self.encoding == 'ZZFeatureMap':
            self._ZZFeatureMap_encoding(self.nqbits,inputs)
        else:
            raise NotImplementedError('Data encoding method not implemented')
        
        return self.apply_ansatz(weights_layers)

    def qnn_layer_QAOA(self,inputs, weights_layers, weights_input):
        """
        Creates the quantum neural network composed of the quantum encoding with QAOA encoding, 
        and a quantum ansatz.

        Args
            inputs (array): Input data
            weight_layers (array): Weights of the ansatz layer
            weight_input (array): Weights of the QAOA quantum encoding

        Returns:
            (QuantumNN): quantum neural network
        """
        # Data encoding
        self._QAOA_encoding(inputs, weights_input)
        return self.apply_ansatz(weights_layers)

        
    def get_shapes(self):
        # Define weight shapes
        if self.ansatz=='circuit_10':
            self.weights_shape = (self.n_layers,self.nqbits, 3) 
        elif self.ansatz=='circuit_6':
            self.weights_shape = (self.nqbits,(3 + self.nqbits)*self.n_layers)
        elif self.ansatz=='circuit_9':
            self.weights_shape = (self.nqbits, self.n_layers) 
        elif self.ansatz=='circuit_15':
            self.weights_shape =  ( self.n_layers, self.nqbits) 
        elif self.ansatz=='circuit_14':
            self.weights_shape = (self.nqbits,4*self.n_layers) 
        elif self.ansatz=='circuit_13':
            self.weights_shape = (self.nqbits,4*self.n_layers) 

    def create_layer(self, type_layer='keras'):
        """
        Creates a Quantum fully connected layer and initializes it.

        Args:
            type_layer (str): Type of quantum layer. It can either be 'keras' or 'torch'

        Returns:
            Quantum layer
        """

        # Define weight input shapes
        if self.encoding=='QAOA': 
            if self.nqbits==1:
                self.weights_input_shape = (2,1) # 2 layers
            elif self.nqbits==2:
                self.weights_input_shape = (2,3)
            else:
                self.weights_input_shape = (2,2*self.nqbits)
            shapes = {"weights_input":self.weights_input_shape, "weights_layers": self.weights_shape} # Define shapes
            self.qnode = qml.QNode(self.qnn_layer_QAOA, self.dev) # Create quantum node
        else:
            shapes = {"weights_layers": self.weights_shape} # Define shapes
            self.qnode = qml.QNode(self.qnn_layer, self.dev)# Create quantum node
        # Create quantum layer
        if type_layer=='keras':
            if isinstance(self.observables, list):
                output_dim = len(self.observables)
            else:
                output_dim = 2**self.nqbits
            qlayer = qml.qnn.KerasLayer(self.qnode, shapes, output_dim=output_dim) # Create keras layer
        elif type_layer=='torch':
            if self.encoding=='QAOA': 
                init_method = {"weights_layers": torch.nn.init.normal_, "weights_input": torch.nn.init.normal_}
            else:
                init_method = {"weights_layers": torch.nn.init.normal_}
            qlayer = qml.qnn.TorchLayer(self.qnode, shapes, init_method=init_method) # Create Torch layer
        else:
            raise NotImplementedError(
                'type_layer should either be keras or torch.')
        return qlayer



import numpy as np
from qiskit.quantum_info import Statevector
import pennylane as qml
from qiskit import QuantumCircuit
        
#################
# Exercises chapter 2
#################

def exercise2_1(vector):
    vector = np.array(vector)
    assert np.isclose(np.abs(vector[0])**2, 1/2),  f"Wrong state, try again. The probability of state |0⟩ should be 0.5."
    assert np.angle(vector[0]) == np.angle(vector[1]), f"Wrong state, try again. The two components should have equal sign."
    assert len(vector)==2, f"Wrong state, try again.  Incorrect vector dimension."


def exercise2_2(vector):
    vector = np.array(vector)
    assert np.isclose(np.abs(vector[0])**2, 1/4),  f"Wrong state, try again. The probability of state |0⟩ should be 0.5."
    assert np.angle(vector[0]) == np.angle(vector[1]), f"Wrong state, try again. The two components should have equal sign."
    assert len(vector)==2, f"Wrong state, try again.  Incorrect vector dimension."

def exercise2_3(vector):
    vector = np.array(vector)
    assert np.isclose(np.abs(vector[0])**2, 1/2), f"Wrong state, try again. The probability of state |0⟩ should be 0.25."
    assert np.isclose(np.abs(np.angle(vector[0]) - np.angle(vector[1])), np.pi), f"Wrong state, try again. The two components should have opposite sign."
    assert len(vector)==2, f"Wrong state, try again.  Incorrect vector dimension."


def exercise2_4(counts):
    total_counts = np.sum(list(counts.values()))
    assert total_counts == 4096, f'Incorrect number of runs.'
    assert np.isclose(counts['00']/total_counts, 1/2, 0.1) and np.isclose(counts['11']/total_counts, 1/2, 0.1), f'Incorrect distribution.'

def exercise2_5(vector):
    vector = np.array(vector)
    assert np.isclose(np.abs(vector[0])**2, 1/2) and np.isclose(np.abs(vector[3])**2, 0.3) and np.isclose(np.abs(vector[7])**2, 0.2), f"Wrong state, try again. The probability of the basis states are incorrect."
    assert np.isclose(np.angle(vector[0]), np.pi/2) and np.isclose(np.angle(vector[3]), np.pi) and np.isclose(np.angle(vector[7]), 0), f"Wrong state, try again. The two components should have opposite sign."
    assert len(vector)==2**3, f"Wrong state, try again.  Incorrect vector dimension."

        
#################
# Exercises chapter 3
#################

def exercise3_1(qc):
    vector = Statevector(qc).data
    assert np.isclose(np.abs(vector[0])**2, 1),  f"Wrong state, try again. The probability of state |0⟩ should be 1."
    assert len(vector)==2, f"Wrong state, try again.  Incorrect vector dimension."

def exercise3_2(qc):
    vector = Statevector(qc).data
    assert np.isclose(np.abs(vector[0b0101])**2, 1),  f"Wrong state, try again. The state should be |0101⟩"
    assert len(vector)==2**4, f"Wrong state, try again.  Incorrect vector dimension."

def exercise3_3(qc):
    vector = Statevector(qc)
    assert len(vector)==4, f"Wrong state, try again.  Incorrect vector dimension."
    assert np.isclose(np.abs(vector[1])**2 + np.abs(vector[3])**2, 0.8),  f"Wrong state, try again. The probability of qubit 1 being on state |1⟩ should be 0.8."
    
def exercise3_4(qc):
    vector = Statevector(qc)
    assert len(vector)==4, f"Wrong state, try again.  Incorrect vector dimension."
    assert np.isclose(np.abs(vector[1])**2, 0.5),  f"Wrong state, try again. The probability of state |01⟩ should be 0.5."
    assert np.isclose(np.abs(vector[2])**2, 0.5),  f"Wrong state, try again. The probability of state |10⟩ should be 0.5."
    assert np.isclose(np.abs(np.angle(vector[1]) - np.angle(vector[2])), np.pi), f"Wrong state, try again. The two components should have opposite sign."

def exercise3_5(qc):
    vector = Statevector(qc)
    assert len(vector)==2**3, f"Wrong state, try again.  Incorrect vector dimension."
    assert np.isclose(np.abs(vector[0])**2, 0.5),  f"Wrong state, try again. The probability of state |000⟩ should be 0.5."
    assert np.isclose(np.abs(vector[7])**2, 0.5),  f"Wrong state, try again. The probability of state |111⟩ should be 0.5."
    assert np.isclose(np.abs(np.angle(vector[0]) - np.angle(vector[7])),0), f"Wrong state, try again. The two components should have the same sign."
    
def exercise3_6(qc):
    gate_names = [op[0].name for op in qc.data]
    for g in gate_names:
        assert g in ['cx', 'x', 'h', 'rx', 'rz', 'ry', 'z'], f"Use the gates you learned in this lesson!"
    assert qc.num_qubits==2, f"Wrong state, try again.  Incorrect vector dimension."
    # Try state 00
    qc_init = QuantumCircuit(2)
    qc_all = qc_init.compose(qc, qubits=[0,1])
    vector = Statevector(qc_all)
    assert np.isclose(np.abs(vector[0])**2, 1),  f"Wrong state, try again. The circuit shoud map the state |00⟩ to the state |00⟩"
    # Try state 01
    qc_init = QuantumCircuit(2)
    qc_init.x(0)
    qc_all = qc_init.compose(qc, qubits=[0,1])
    vector = Statevector(qc_all)
    assert np.isclose(np.abs(vector[2])**2, 1),  f"Wrong state, try again. The circuit shoud map the state |01⟩ to the state |10⟩"
    # Try state 01
    qc_init = QuantumCircuit(2)
    qc_init.x(1)
    qc_all = qc_init.compose(qc, qubits=[0,1])
    vector = Statevector(qc_all)
    assert np.isclose(np.abs(vector[1])**2, 1),  f"Wrong state, try again. The circuit shoud map the state |10⟩ to the state |01⟩"
    # Try state 11
    qc_init = QuantumCircuit(2)
    qc_init.x(0)
    qc_init.x(1)
    qc_all = qc_init.compose(qc, qubits=[0,1])
    vector = Statevector(qc_all)
    assert np.isclose(np.abs(vector[3])**2, 1),  f"Wrong state, try again. The circuit shoud map the state |11⟩ to the state |11⟩"

def get_secret_state(theta=np.pi, phi = np.pi/2):
    qc = QuantumCircuit(1)
    qc.ry(2 * theta, 0)
    qc.rz(phi, 0)
    return qc

def exercise3_7(qc):
    vector = Statevector(qc)
    assert np.isclose(np.abs(vector[0])**2, 1),  f"Wrong state, try again. The circuit should output state |0⟩"
    
def exercise3_8(qc, phi):
    vector = Statevector(qc)
    assert len(vector)==2, f"Wrong state, try again.  Incorrect vector dimension."
    assert np.isclose(np.abs(vector[0])**2, 0.5),  f"Wrong state, try again. The probability of state |0⟩ should be 0.5."
    assert np.isclose(np.abs(vector[1])**2, 0.5),  f"Wrong state, try again. The probability of state |1⟩ should be 0.5."
    assert np.isclose(np.abs(np.angle(vector[0]) - np.angle(vector[1])),phi), f"Wrong state, try again. The relative phase should be {phi:.5f}"
    
def exercise3_9(result_x, result_y, phi):
    total_counts = np.sum(list(result_x.values()))
    p_plus = result_x['0']/total_counts
    p_iplus = result_y['0']/total_counts
    assert np.isclose(p_plus,(1 + np.cos(phi))/2, atol=0.1), f'Incorrect measurement in the X basis.'
    assert np.isclose(p_iplus,(1 + np.sin(phi))/2, atol=0.1), f'Incorrect measurement in the X basis.'

def exercise3_10(phi_reconstructed, phi):
    assert np.isclose(phi_reconstructed, phi, atol=0.1), f'Incorrect value of phi = {phi_reconstructed}'

#################
# Exercises chapter 4
#################


def exercise4_1(num_qubits, num_gates, depth, gates):
    print(f"Number of qubits: {num_qubits}, Number of gates: {num_gates}, Depth: {depth}, Gates: {gates}")
    assert num_qubits==4, f'Wrong number of qubits'
    assert num_gates==80, f'Wrong number of gates'
    assert depth==41, f'Wrong depth'
    assert gates=={'Rot':40, 'CNOT':40}, f'Wrong number of qubits'    


def exercise4_3(qnode, num_qubits, num_gates, depth, X_train, random_weights):
    print(f"Number of qubits: {num_qubits}, Number of gates: {num_gates}, Depth: {depth}")
    num_qubits2 = []
    num_gates2 = []
    depth2 = []
    for X in X_train[:10]:
        qnode(X, random_weights)
        expanded_tape = qnode.qtape.expand(depth=2)
        specs = expanded_tape.specs
        num_qubits2.append(specs['resources'].num_wires)
        num_gates2.append(specs['resources'].num_gates)
        depth2.append(specs['resources'].depth)
    num_qubits2 = np.mean(num_qubits2)
    num_gates2 = np.mean(num_gates2)
    depth2 = np.mean(depth2)
    assert num_qubits==num_qubits2, f'Wrong number of qubits'
    assert num_gates==num_gates2, f'Wrong number of gates'
    assert depth==depth2, f'Wrong depth'

#################
# Exercises chapter 5 and 6
#################
def exercise5_7(total_single_qubit_gates, total_entangling_gates, processor):
    total_single_qubit_gates2 = processor.info['num_circuits']*(processor.info['gates']['RX'] + processor.info['gates']['H'])
    total_entangling_gates2 = processor.info['num_circuits']*processor.info['num_nonlocal_gates']
    assert total_single_qubit_gates==total_single_qubit_gates2, f'Wrong total number of single qubit gates.'
    assert total_entangling_gates2==total_entangling_gates, f'Wrong total number of entangling qubit gates.'


def exercise6_1(n_blocks, nqbits=8):
    true_nblocks = qml.MERA.get_n_blocks(range(nqbits), 2)
    assert true_nblocks==n_blocks

def exercise6_2(template_weights):
    assert np.array(template_weights).shape==(3,2,4)

def exercise6_3(gates):
    assert gates==[84, 180, 756, 1524], f'Wrong number of gates.'

def exercise6_4(params_model_reduced):
    assert params_model_reduced==7162792, f'Wrong number of parameters for compressed model'

def exercise6_6(params_model_reduced):
    assert params_model_reduced==7897351, f'Wrong number of parameters for compressed model'


import re
from setuptools import setup

with open("README.md", "r") as fh:
    readme = fh.read()

regex = r"^Version:\s(\d+.\d+.\d+)$"
match = re.search(regex, readme, re.MULTILINE)

if match:
    current_version = match.group(1)
else:
    raise Exception("Current version is missing from the README.md")

with open("requirements.txt", "r") as r:
    requirements = [p for p in r.read().split("\n") if p]
    dependencies = [p.split("=")[0] for p in requirements]

setup(
    version=current_version,
    package_data={
        "ingenii_quantum": ["data/*"],  # Include all files in the data folder
    },
    include_package_data=True,  # This is required for package_data to work
    install_requires=requirements,
    dependency_links=dependencies
)



from pennylane import numpy as np
from pennylane.math import partial_trace, reduce_statevector, fidelity_statevector
from scipy.special import kl_div


class EntanglingCapacity:
    def __init__(self, circuit, dev, params_shape):
        """
        Initializes the EntanglingCapacity class with the given circuit, device, and parameter shape.

        Args:
            circuit: the QNode representing the quantum circuit
            dev: quantum device used for simulation
            params_shape: shape of the parameters for the circuit
        """
        self.circuit = circuit
        self.dev = dev
        self.params_shape = params_shape
        self.N = len(dev.wires)  # Number of qubits

    def generate_random_params(self):
        """
        Generates random parameters for the quantum circuit based on the specified parameter shape.
        """
        return np.random.uniform(0, 2 * np.pi, self.params_shape)

    def meyer_wallach_entanglement(self, n_samples):
        """
        Computes the Meyer-Wallach entanglement measure for the quantum circuit.

        Args:
            n_samples (int): Number of samples to calculate the entanglement measure

        Returns:
            float: Meyer-Wallach entanglement measure averaged over multiple samples.
        """
        res = np.zeros(n_samples, dtype=complex)

        for i in range(n_samples):
            params = self.generate_random_params()  # Random parameters for the PQC
            state = self.circuit(params)  # Get quantum state from the PQC

            # Reduce the full state vector to density matrix for the entire system
            rho = reduce_statevector(state, indices=self.dev.wires)

            entropy = self._calculate_entropy(rho)

            # Meyer-Wallach measure for the current state
            res[i] = 1 - (entropy / self.N)

        # Average over the samples and return
        return float(2 * np.mean(res).real)

    def _calculate_entropy(self, rho):
        """
        Helper function to calculate the average entropy over all qubits.

        Args:
            rho: the reduced state of the quantum system (density matrix).
            
        Returns:
            float: Entropy
        """
        entropy = 0
        qb_indices = list(range(self.N))

        # Loop over each qubit, calculate the partial trace and its entropy
        for j in range(self.N):
            # Partial trace over all qubits except the j-th qubit
            dens = partial_trace(rho, qb_indices[:j] + qb_indices[j+1:])
            trace = np.trace(dens**2)  # Calculate the purity (trace of the square of density matrix)
            entropy += trace

        return entropy

class Expressibility:
    def __init__(self, pqc, params_shape, dev):
        """
        Initialize the Expressibility class.

        Args:
            pqc (qml.QNode): Parameterized quantum circuit.
            params_shape (tuple): Shape of the parameters used in the PQC.
            dev (qml.Device): Pennylane device for executing the PQC.
        """
        self.pqc = pqc       
        self.params_shape = params_shape
        self.dev = dev

    def generate_random_params(self, params_shape):
        """
        Generate random parameters for the parameterized quantum circuit.

        Args:
            params_shape (tuple): Shape of the parameters.

        Returns:
            np.ndarray: Randomly generated parameters in the range [0, 2π].
        """
        return np.random.uniform(0, 2 * np.pi, params_shape)

    def pqc_fidelity(self, n_samples):
        """
        Calculate the fidelity between quantum states produced by the PQC.

        Args:
            n_samples (int): Number of samples for fidelity computation.

        Returns:
            list: List of fidelity values for the sampled states.
        """
        fidelities = []
        for _ in range(n_samples):
            # Get random parameters for PQC1 and PQC2
            params1 = self.generate_random_params(self.params_shape)
            params2 = self.generate_random_params(self.params_shape)
            state1 = self.pqc(params1)
            state2 = self.pqc(params2)
            # Calculate fidelity between the two states
            fidelities.append(fidelity_statevector(state1, state2))
        return fidelities

    def haar_fidelity(self, n_samples):
        """
        Generate fidelities for Haar-random states.

        Args:
            n_samples (int): Number of samples for fidelity computation.

        Returns:
            np.ndarray: Array of Haar-random fidelities.
        """
        n_qubits = len(self.dev.wires)
        N = 2**n_qubits
        F = np.random.uniform(0, 1, n_samples)
        return (N-1) * (1 - F)**(N-2)

    def compute_expressibility(self, n_samples):
        """
        Compute the expressibility of the PQC by comparing its fidelity distribution
        to that of Haar-random states using KL divergence.

        Args:
            n_samples (int): Number of samples for the computation.

        Returns:
            float: Expressibility value, computed as the mean KL divergence.
        """
        pqc_f = self.pqc_fidelity(n_samples)
        haar_f = self.haar_fidelity(n_samples)
        return float(np.mean(kl_div(pqc_f, haar_f)))
